﻿using System;
using System.Xml;
using System.Linq;
using System.Text;
using System.Data;
using DemoAppSQLUtils;
using WebReports.Api;
using System.Data.SqlClient;
using WebReports.Api.Common;
using System.Collections.Generic;

namespace ExagoExtensibility
{
	/// <summary>
	/// This represents the implementation of a specific set of methods, as defined by Exago's Folder Management 
	/// itself, such that, if configured in Exago's Admin Console to use Folder Management, these methods will be 
	/// referenced for the storage and retrieval of the many aspects of Exago report definitions.
	/// </summary>
    class FolderManagement
    {
        private static char[] tokens = { '\\', '/' };	// folder and report name parsing delimiters

		/**
		 * These differentiate the various types of objects that are saved via Folder Management.  Note
		 * that reports and folders are both saved as ReportType = 0.
		 */
		public static class TypeIds
        {
            public static int ReportType = 0;
            public static int TemplateType = 1;
            public static int ThemeType = 2;
        }

		/**
		 * Any XML document node, for the purpose of rebuilding an XML document from database records
		 */ 
        public class Entity
        {
            public int Id;
            public string Name;
            public bool LeafFlag;
            public bool ReadOnlyFlag;
            public string Description;
            public List<Entity> Children = new List<Entity>();

            public Entity(int id, string name, bool leafFlag,  bool readOnlyFlag, string description)
            {
                this.Id = id;
                this.Name = name;
                this.LeafFlag = leafFlag;
                this.ReadOnlyFlag = readOnlyFlag;
                this.Description = description;
            }
        }

		/**
		 * Return XML representing the hierarchy of reports and folders that the active user has access to. The 
		 * active user is determined by userId.  The permitted access comes from three database tables providing 
		 * user, role and public level control.
		 */ 
        public static string GetReportListXml(SessionInfo sessionInfo)
        {
            List<Entity> entities = new List<Entity>();
            Dictionary<int, Entity> entityIdMapping = new Dictionary<int, Entity>();
			SQLHandler sql = new SQLHandler();

			sql.Stmt = string.Format(
				"SELECT id, name, content, leaf_flag, bit_content, type, themeType, parentId, " +
					"MIN( CASE WHEN readonly_flag=1 THEN 1 ELSE 0 END ) AS readonly_flag " +
				"FROM (" +
					"SELECT {0}.id, {0}.name, {0}.content, {0}.leaf_flag, {0}.bit_content, {0}.type, {0}.isDeleted, " +
						"{0}.themeType, {0}.parentId, {1}.readonly_flag " +
					"FROM {0} "+
					"INNER JOIN {1} ON {0}.id={1}.reportId " +
					"WHERE {1}.userId=@userId " +
					"AND {0}.type={5} " +
					"AND {0}.isDeleted='false' " +
					"UNION " +
					"SELECT {0}.id, {0}.name, {0}.content, {0}.leaf_flag, {0}.bit_content, {0}.type, {0}.isDeleted, " +
						"{0}.themeType, {0}.parentId, {2}.readonly_flag " +
					"FROM {2} " +
					"INNER JOIN {0} on {0}.id={2}.reportId " +
					"INNER JOIN {3} on {3}.classId={2}.userClassId " +
					"WHERE {3}.userId=@userId " +
					"AND {0}.type={5} " +
					"AND {0}.isDeleted='false' " +
					"UNION " +
					"SELECT {0}.id, {0}.name, {0}.content, {0}.leaf_flag, {0}.bit_content, {0}.type, {0}.isDeleted, " +
						"{0}.themeType, {0}.parentId, {4}.readonly_flag " +
					"FROM {0} " +
					"INNER JOIN {4} ON {0}.id={4}.reportId " +
					"WHERE {0}.type={5} " +
					"AND {0}.isDeleted='false') " +
				"AS ReportListXML " +
				"GROUP BY id, name, content, bit_content, type, leaf_flag, themeType, parentId", 
				DbTables.Reports, DbTables.RptAccessUser, DbTables.RptAccessUserClasses, 
				DbTables.UserSettings, DbTables.RptPublished, TypeIds.ReportType);
            sql.SqlParams.Add(new SqlParameter("@userId", sessionInfo.GetParameter("userId").Value));
            DataSet ds = sql.ExecuteReadCmd();

			// The idea is to loop through the returned dataset adding nodes, until no more can be added.  Since a child may appear
			// in the dataset before its parent and the parent needs to exist in the XmlDoc before adding its child, we can't
			// just loop through the list and add all nodes sequentially.
			List<DataRow> deleteList = new List<DataRow>();
			bool nodeAdded = false;
			while (true)
            {
                nodeAdded = false;

                if (ds == null)
                    return "<entity></entity>";

                foreach (DataRow dr in ds.Tables[0].Rows)
                {
                    string name = (string)dr["name"];
                    int id = (int)dr["id"];
                    bool leafFlag = ((bool)dr["leaf_flag"]) == true ? true : false;
                    bool readOnlyFlag = ((int)dr["readonly_flag"]) == 1 ? true : false;
                    object parentObj = dr["parentId"];

                    if (parentObj.Equals(DBNull.Value))		// This entity is a root folder
					{
                        entities.Add(new Entity(id, name, false, readOnlyFlag, null));
                        entityIdMapping.Add(id, entities[entities.Count - 1]); // Keep a "pointer" to the node so we can easily access it by ID when we need to add children

                        nodeAdded = true;
                        deleteList.Add(dr);
                    }
					else                                    // The entity is a child of a folder.  Find the parent entity and add this entity as a child.
					{
                        string parent_id = "" + (int)parentObj;

                        if (!entityIdMapping.ContainsKey((int)parentObj))
                            continue;

                        Entity parent = entityIdMapping[(int)parentObj];

                        nodeAdded = true;
                        deleteList.Add(dr);

                        string description = null;

                        if (dr["content"].GetType() != typeof(DBNull))
                        {
                            XmlDocument reportDoc = new XmlDocument();
                            reportDoc.LoadXml((string)dr["content"]);
                            XmlNode descriptionNode = reportDoc.SelectSingleNode("/report/main/description");

                            if (descriptionNode != null)
                                description = descriptionNode.InnerText;
                        }

                        parent.Children.Add(new Entity(id, name, leafFlag, readOnlyFlag, description));
                        entityIdMapping.Add(id, parent.Children[parent.Children.Count - 1]);
                    }
                }

                foreach (DataRow dr in deleteList)
                    ds.Tables[0].Rows.Remove(dr);

                if (!nodeAdded && ds.Tables[0].Rows.Count > 0)
                    throw new Exception("Table malformatted: Found child with nonexistent parent");
                else if (!nodeAdded)
                    break;

                deleteList.Clear();
            }

            // Recursively sort the entities list (Reports before folders; everything alphabetically)
            FolderManagement.SortEntities(entities);

            XmlDocument xDoc = new XmlDocument();
            XmlNode rootNode = xDoc.CreateElement("eWebReportsReports");
            xDoc.AppendChild(rootNode);

			// Render the entities list as XML to return to the caller
			string folderxml = FolderManagement.WriteXml(entities, xDoc, rootNode);
            return folderxml;
        }

		/**
		 * Sort the entities list such that reports come before folders and everything is sorted
		 * alphabetically.
		 */ 
		private static void SortEntities(List<Entity> entities)
        {
            foreach (Entity entity in entities)
            {
                entities.Sort((a, b) =>
                {
                    if (a.LeafFlag && !b.LeafFlag)
                        return -1;
                    if (b.LeafFlag && !a.LeafFlag)
                        return 1;
                    return a.Name.CompareTo(b.Name);
                });

                SortEntities(entity.Children);
            }
        }

		/**
		 * Write a list of entities out as an XML document
		 */ 
        private static string WriteXml(List<Entity> entities, XmlDocument xmlDoc, XmlNode root)
        {
            foreach (Entity entity in entities)
            {
                XmlNode node = xmlDoc.CreateElement("entity");
                root.AppendChild(node);

                XmlAttribute attribute = xmlDoc.CreateAttribute("id");
                attribute.Value = entity.Id.ToString();
                node.Attributes.Append(attribute);
                XmlNode nameNode = xmlDoc.CreateElement("name");
                node.AppendChild(nameNode);
                nameNode.InnerText = entity.Name;
                XmlNode leafNode = xmlDoc.CreateElement("leaf_flag");
                node.AppendChild(leafNode);
                leafNode.InnerText = entity.LeafFlag.ToString();
                XmlNode roNode = xmlDoc.CreateElement("readonly_flag");
                node.AppendChild(roNode);
                roNode.InnerText = entity.ReadOnlyFlag.ToString();

                if (entity.Description != null)
                {
                    XmlNode descriptionNode = xmlDoc.CreateElement("description");
                    node.AppendChild(descriptionNode);
                    descriptionNode.InnerText = entity.Description;
                }

                WriteXml(entity.Children, xmlDoc, node);
            }

            return xmlDoc.InnerXml;
        }

		/*
		 *  Add a report folder (folderName is fully qualified). Folder should not be named to one that already exists.
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param folderName is a fully-qualified folder name
		 *  @return an error message string or, if successful, null
		 */
		public static string AddFolder(SessionInfo sessionInfo, string folderName)
        {
            if (ExistFolder(sessionInfo, folderName))
                return "A folder with that name already exists."; // TODO:  should return a language-based message from the exago language xml file

			string[] pathParts = folderName.Split(tokens);
            string parentIdStr = GetParentFolderIdString(sessionInfo, pathParts, false);

			Dictionary<string, object> fldrInfo = new Dictionary<string, object>();
			fldrInfo.Add("name", pathParts[pathParts.Length - 1]);
			fldrInfo.Add("parent", parentIdStr);
			return AddRptItem(sessionInfo, TypeIds.ReportType, fldrInfo);
        }

		/*
		 *  Delete a report folder.
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param folderName is a fully-qualified folder name
		 *  @return an error message string or, if successful, null
		 */
		public static string DeleteFolder(SessionInfo sessionInfo, string folderName)
        {
            if (folderName == "My Reports")
                return "Cannot delete the folder 'My Reports'";//root folder

			string[] pathParts = folderName.Split(tokens);
            string folderIdStr = GetFolderIdString(sessionInfo, pathParts, true);

			Dictionary<string, object> fldrInfo = new Dictionary<string, object>();
			fldrInfo.Add("change", "isDeleted='true'");
			fldrInfo.Add("where", "id " + folderIdStr);
			return UpdateRptItem(sessionInfo, fldrInfo);
        }

		/*
		 *  Rename a report folder. Folder cannot be renamed to one that already exists.
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param oldName is a fully-qualified, pre-existing folder name
		 *  @param newName is a fully-qualified, non-existing folder name
		 *  @return an error message string or, if successful, null
		 */
		public static string RenameFolder(SessionInfo sessionInfo, string oldName, string newName)
        {
            if (oldName == "My Reports")
                return "Cannot change the name of 'My Reports'";

			string[] oldPathParts = oldName.Split(tokens);
            string[] newPathParts = newName.Split(tokens);
            string folderIdStr = GetFolderIdString(sessionInfo, oldPathParts, true);

			Dictionary<string, object> fldrInfo = new Dictionary<string, object>();
			fldrInfo.Add("change", "name='" + newPathParts[newPathParts.Length - 1].Replace("'", "''") + "'");
			fldrInfo.Add("where", "id " + folderIdStr);
			return UpdateRptItem(sessionInfo, fldrInfo);
		}

		/*
		 *  Rename a report folder to a new parent. Folder should not be moved to a folder where the moved folder name already exists.
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param oldName is a fully-qualified, pre-existing folder name
		 *  @param newName is a fully-qualified, non-existing folder name
		 *  @return an error message string or, if successful, null
		 */
		public static string MoveFolder(SessionInfo sessionInfo, string oldName, string newName)
        {
            if (oldName == "My Reports")
                return "Cannot move the folder 'My Reports'";

			string[] oldPathParts = oldName.Split(tokens);
            string[] newPathParts = newName.Split(tokens);
            string folderIdStr = GetFolderIdString(sessionInfo, oldPathParts, true);
            string newParentfolderIdStr = GetParentFolderIdString(sessionInfo, newPathParts, false);

			Dictionary<string, object> fldrInfo = new Dictionary<string, object>();
			fldrInfo.Add("change", "parentId=" + newParentfolderIdStr);
			fldrInfo.Add("where", "id " + folderIdStr);
			return UpdateRptItem(sessionInfo, fldrInfo);
        }

		/*
		 *  Determines if a folder exists; returns true or false
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param folderName is a fully-qualified, pre-existing folder name
		 *  @return true if folder exists and false otherwise
		 */
		public static bool ExistFolder(SessionInfo sessionInfo, string folderName)
        {
            string[] pathParts = folderName.Split(tokens);
            string parentIdStr = GetParentFolderIdString(sessionInfo, pathParts, true);

			Dictionary<string, object> rptInfo = new Dictionary<string, object>();
			rptInfo.Add("format", "single");
			rptInfo.Add("whatToGet", "name");
			rptInfo.Add("type", TypeIds.ReportType.ToString());
			rptInfo.Add("parent", parentIdStr);
			rptInfo.Add("name", pathParts[pathParts.Length - 1]);
			rptInfo.Add("leaf", "false");
			return (GetRptItem(sessionInfo, rptInfo) != null);
        }

		/*
		 *  Gets a report
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param reportName is a fully-qualified, pre-existing report name
		 *  @return an error message string or, if successful, the report xml string
		 */
		public static string GetReportXml(SessionInfo sessionInfo, string reportName)
        {
            string[] pathParts = reportName.Split(tokens);
            string parentIdStr = GetParentFolderIdString(sessionInfo, pathParts, true);

			Dictionary<string, object> rptInfo = new Dictionary<string, object>();
			rptInfo.Add("format", "single");
			rptInfo.Add("whatToGet", "content");
			rptInfo.Add("type", TypeIds.ReportType.ToString());
			rptInfo.Add("parent", parentIdStr);
			rptInfo.Add("name", pathParts[pathParts.Length - 1]);
			rptInfo.Add("leaf", "true");
			return (string)GetRptItem(sessionInfo, rptInfo);
        }

		/*
		 *  Determines if a report currently exists
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param reportName is a fully-qualified, pre-existing report name
		 *  @return the report id, if it exists; otherwise, null
		 */
		public static string ExistReport(SessionInfo sessionInfo, string reportName)
        {
            string[] pathParts = reportName.Split(tokens);
            string parentIdStr = GetParentFolderIdString(sessionInfo, pathParts, true);

			Dictionary<string, object> rptInfo = new Dictionary<string, object>();
			rptInfo.Add("format", "single");
			rptInfo.Add("whatToGet", "id");
			rptInfo.Add("type", TypeIds.ReportType.ToString());
			rptInfo.Add("parent", parentIdStr);
			rptInfo.Add("name", pathParts[pathParts.Length - 1]);
			rptInfo.Add("leaf", "true");
			object obj = GetRptItem(sessionInfo, rptInfo);
			if (obj == null)
				return null;

			return obj.ToString();
        }

		/*
		 *  Save a report
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param reportName is a fully-qualified, pre-existing report name
		 *  @param reportXml is the XML string representation of the report
		 *  @return an error message string or, if successful, null
		 */
		public static string SaveReport(SessionInfo sessionInfo, string reportName, string reportXml)
        {
            //check if report with same name already exists
            string reportId = ExistReport(sessionInfo, reportName);

            if (reportId == null)
            {
                string[] pathParts = reportName.Split(tokens);
                string reportFolderIdString = GetParentFolderIdString(sessionInfo, pathParts, false);

				Dictionary<string, object> rptInfo = new Dictionary<string, object>();
				rptInfo.Add("name", pathParts[pathParts.Length - 1]);
				rptInfo.Add("report", reportXml);
				rptInfo.Add("parent", reportFolderIdString);
				return AddRptItem(sessionInfo, TypeIds.ReportType, rptInfo);
            }
            else
            {
				Dictionary<string, object> rptInfo = new Dictionary<string, object>();
				rptInfo.Add("change", "content='" + reportXml.Replace("'", "''") + "'");
				rptInfo.Add("where", "id=" + reportId);
				return UpdateRptItem(sessionInfo, rptInfo);
            }
        }

		/*
		 *  Duplicates a report. If this method is not provided, SaveReport will be called.
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param reportName is a fully-qualified, pre-existing report name
		 *  @param reportXml is the XML string representation of the report
		 *  @return an error message string or, if successful, null
		 */
		public static string DuplicateReport(SessionInfo sessionInfo, string reportName, string reportXml)
        {
            return SaveReport(sessionInfo, reportName, reportXml);
        }

		/*
		 *  Deletes a report (not really; just marked as deleted).
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param reportName is a fully-qualified, pre-existing report name
		 *  @return an error message string or, if successful, null
		 */
		public static string DeleteReport(SessionInfo sessionInfo, string reportName)
        {
            string reportId = ExistReport(sessionInfo, reportName);

			Dictionary<string, object> rptInfo = new Dictionary<string, object>();
			rptInfo.Add("change", "isDeleted='true'");
			rptInfo.Add("where", "id=" + reportId);
			return UpdateRptItem(sessionInfo, rptInfo);
        }

		/*
		 *  Renames a report.
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param reportName is a fully-qualified, pre-existing report name
		 *  @param newReportName is a fully-qualified, non-existing report name
		 *  @return an error message string or, if successful, null
		 */
		public static string RenameReport(SessionInfo sessionInfo, string reportName, string newReportName)
        {
            string[] newPathParts = newReportName.Split(tokens);
            string newParentIdStr = GetParentFolderIdString(sessionInfo, newPathParts, true);
            string reportId = ExistReport(sessionInfo, reportName);

			Dictionary<string, object> rptInfo = new Dictionary<string, object>();
			rptInfo.Add("change", "name='" + newPathParts[newPathParts.Length - 1].Replace("'", "''") + "', parentId " + newParentIdStr);
			rptInfo.Add("where", "id=" + reportId);
			return UpdateRptItem(sessionInfo, rptInfo);
        }

		/*
		 *  Gets a list of the filenames of all the templates currently stored
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @return a list of template filenames (strings)
		 */
		public static List<string> GetTemplateList(SessionInfo sessionInfo)
        {
			Dictionary<string, object> tmplInfo = new Dictionary<string, object>();
			tmplInfo.Add("format", "list");
			tmplInfo.Add("whatToGet", "name");
			tmplInfo.Add("type", TypeIds.TemplateType.ToString());
			return (List<string>)GetRptItem(sessionInfo, tmplInfo);
		}

		/*
		 *  Gets a byte array representation of a template
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param templateName is the name of the template requested
		 *  @return a byte array representation (binary content) of a template
		 */
		public static byte[] GetTemplate(SessionInfo sessionInfo, string templateName)
        {
			Dictionary<string, object> tmplInfo = new Dictionary<string, object>();
			tmplInfo.Add("format", "single");
			tmplInfo.Add("whatToGet", "bit_content");
			tmplInfo.Add("type", TypeIds.TemplateType.ToString());
			tmplInfo.Add("name", templateName);
			return (byte[])GetRptItem(sessionInfo, tmplInfo);
        }

		/*
		 *  Saves a byte array representation of a template
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param templateName is the name of the template requested
		 *  @param templateData is the binary/byte array representation of the template
		 *  @return an error message string or, if successful, null
		 */
		public static string SaveTemplate(SessionInfo sessionInfo, string templateName, byte[] templateData)
        {
			Dictionary<string, object> tmplInfo = new Dictionary<string, object>();
			tmplInfo.Add("format", "single");
			tmplInfo.Add("whatToGet", "name");
			tmplInfo.Add("type", TypeIds.TemplateType.ToString());
			tmplInfo.Add("name", templateName);
			object chkObj = GetRptItem(sessionInfo, tmplInfo);

			if (chkObj != null)
                return "Error saving template file: File exists";

			tmplInfo.Clear();
			tmplInfo.Add("name", templateName);
			tmplInfo.Add("template", templateData);
			return AddRptItem(sessionInfo, TypeIds.TemplateType, tmplInfo);
        }

		/*
		 *  Gets a list of the filenames of all the themes currently stored
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param themeType is the type qualifier of themes to be retrieved
		 *  @return a list of theme filenames
		 */
		public static List<string> GetThemeList(SessionInfo sessionInfo, string themeType)
        {
			Dictionary<string, object> themeInfo = new Dictionary<string, object>();
			themeInfo.Add("format", "list");
			themeInfo.Add("whatToGet", "name");
			themeInfo.Add("type", TypeIds.ThemeType.ToString());
			themeInfo.Add("themeType", themeType);
			return (List<string>)GetRptItem(sessionInfo, themeInfo);
		}

		/*
		 *  Determines if a theme currently exists
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param themeType is the type qualifier of themes to check
		 *  @param themeName is a fully-qualified, pre-existing theme name
		 *  @return true if the theme exists; otherwise, false
		 */
		public static bool ExistTheme(SessionInfo sessionInfo, string themeType, string themeName)
        {
			Dictionary<string, object> themeInfo = new Dictionary<string, object>();
			themeInfo.Add("format", "single");
			themeInfo.Add("whatToGet", "name");
			themeInfo.Add("type", TypeIds.ThemeType.ToString());
			themeInfo.Add("name", themeName);
			themeInfo.Add("themeType", themeType);
			return (GetRptItem(sessionInfo, themeInfo) != null);
        }

		/*
		 *  Gets a theme
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param themeType is the type qualifier of themes to retrieve
		 *  @param themeName is a fully-qualified, pre-existing theme name
		 *  @return a XML string representation of the theme or a default "theme" if it doesn't exist
		 */
		public static string GetThemeXml(SessionInfo sessionInfo, string themeType, string themeName)
        {
			Dictionary<string, object> themeInfo = new Dictionary<string, object>();
			themeInfo.Add("format", "single");
			themeInfo.Add("whatToGet", "content");
			themeInfo.Add("type", TypeIds.ThemeType.ToString());
			themeInfo.Add("name", themeName);
			themeInfo.Add("themeType", themeType);
			object obj = GetRptItem(sessionInfo, themeInfo);

			if (obj == null)
                return "#942511, #c83a16, #e77200, #f0bf00, #c38307"; //WHY???

            return obj.ToString();
        }

		/*
		 *  Saves a theme
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param themeType is the type qualifier of themes to retrieve
		 *  @param themeName is a fully-qualified, pre-existing theme name
		 *  @param themeXml is the theme content itself to be saved
		 *  @return a string error message or, if successful, null
		 */
		public static string SaveTheme(SessionInfo sessionInfo, string themeType, string themeName, string themeXml)
        {
			Dictionary<string, object> themeInfo = new Dictionary<string, object>();
			themeInfo.Add("format", "single");
			themeInfo.Add("whatToGet", "id");
			themeInfo.Add("type", TypeIds.ThemeType.ToString());
			themeInfo.Add("name", themeName);
			themeInfo.Add("themeType", themeType);
			object obj = GetRptItem(sessionInfo, themeInfo);

			if (obj == null)
			{
				themeInfo.Clear();
				themeInfo.Add("name", themeName);
				themeInfo.Add("theme", themeXml);
				themeInfo.Add("type", themeType);
				return AddRptItem(sessionInfo, TypeIds.ThemeType, themeInfo);
            }
            else
            {
				themeInfo.Clear();
				themeInfo.Add("change", "content='" + themeXml.Replace("'", "''") + "'");
				themeInfo.Add("where", "id=" + obj.ToString());
				return UpdateRptItem(sessionInfo, themeInfo);
            }
        }

		/*
		 *  Get a *parent* folder id based on a fully-qualified folder name
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param pathParts is an array of the folder path names
		 *  @param isQuery is a flag denoting that this will or will not be part of a SELECT (true) or UPDATE/INSERT (false)
		 *  @return a string folder id
		 */
		private static string GetParentFolderIdString(SessionInfo sessionInfo, string[] pathParts, bool isQuery)
        {
            return GetFolderIdString(sessionInfo, pathParts.Take(pathParts.Length - 1).ToArray(), isQuery);
        }

		/*
		 *  Get a folder id based on a fully-qualified folder name
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param pathParts is an array of the folder path names
		 *  @param isQuery is a flag denoting that this will or will not be part of a SELECT (true) or UPDATE/INSERT (false)
		 *  @return a string folder id
		 */
		private static string GetFolderIdString(SessionInfo sessionInfo, string[] pathParts, bool isQuery)
        {
            int? folderId = GetFolderId(sessionInfo, pathParts);
            if (folderId == null) //root folder
            {
                if (isQuery)
                    return "is NULL";
                else
                    return "NULL";
            }
            else
            {
                if (isQuery)
					return "=" + folderId;
				else
					return folderId + "";
			}
		}

		/*
		 *  Iterate through a folder path and get a folder id of a parent folder
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param pathParts is an array of the folder path names
		 *  @return a folder id or null if none exists
		 */
		private static int? GetFolderId(SessionInfo sessionInfo, string[] pathParts)
        {
			int? retVal = null;
			Dictionary<string, object> fldrInfo = new Dictionary<string, object>();

			for (int i = 0; i < pathParts.Length; i++)
			{
				fldrInfo.Clear();
				fldrInfo.Add("format", "single");
				fldrInfo.Add("whatToGet", "id");
				fldrInfo.Add("type", TypeIds.ReportType.ToString());
				fldrInfo.Add("parent", (i == 0 ? "IS NULL" : "=" + retVal.ToString()));
				fldrInfo.Add("name", pathParts[i]);
				fldrInfo.Add("leaf", "false");
				retVal = (int)GetRptItem(sessionInfo, fldrInfo);
			}
			return retVal;
        }

		/*
		 *  Generic method for adding a variety of report objects to the Reports table
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param type is a TypeIds type representing what is being added
		 *  @param options is a Dictionary<string, object> object containing parameters necessary for the record insertion
		 *  @return a string error message or, if successful, null
		 */
		private static string AddRptItem(SessionInfo sessionInfo, int type, Dictionary<string, object> options)
		{
			// default values
			object name = options["name"];
			object parentId = DBNull.Value;
			object content = DBNull.Value;
			object leaf_flag = DBNull.Value;
			object bit_content = DBNull.Value;
			object themeType = DBNull.Value;
			
			// based on type of insert, set variables accordingly
			if (type == TypeIds.ReportType)		// switch would be convenient here if it syntactically worked with TypeIds...
			{
				parentId = ((string)options["parent"] == "NULL" ? DBNull.Value : options["parent"]);
				leaf_flag = false;
				if (options.ContainsKey("report"))	// is this a report or a folder?
				{
					content = options["report"];
					leaf_flag = true;
				}
			}
			else if (type == TypeIds.TemplateType)
			{
				bit_content = options["template"];
			}
			else if (type == TypeIds.ThemeType)
			{
				content = options["theme"];
				themeType = options["type"];
			}

			try
			{
				SQLHandler sql = new SQLHandler();
				sql.Stmt = string.Format("INSERT INTO {0} (name, content, leaf_flag, bit_content, type, isDeleted, " +
					"themeType, parentId, dateCreated, createdByUserId, dateLastModified, lastModifiedByUserId) " +
					"VALUES (@name, @content, @lf, CONVERT(varbinary(max), @bc), @type, @deleted, @tt, @pId, GetDate(), " +
					"@userId, GetDate(), @userId); SELECT SCOPE_IDENTITY();", DbTables.Reports);
				sql.SqlParams.Add(new SqlParameter("@name", name));
				sql.SqlParams.Add(new SqlParameter("@type", type));
				sql.SqlParams.Add(new SqlParameter("@deleted", "false"));
				sql.SqlParams.Add(new SqlParameter("@userId", sessionInfo.UserId));
				sql.SqlParams.Add(new SqlParameter("@bc", bit_content));
				sql.SqlParams.Add(new SqlParameter("@pId", parentId));
				sql.SqlParams.Add(new SqlParameter("@tt", themeType));
				sql.SqlParams.Add(new SqlParameter("@lf", leaf_flag));
				sql.SqlParams.Add(new SqlParameter("@content", content));
				// NOTE: SCOPE_IDENTITY retrieves the last id that was inserted into an IDENTITY field

				//since you created the folder, you should have permission to see/edit that folder
				string id = sql.ExecuteScalarCmd().ToString(); // perform insert and get new id queried from SCOPE_IDENTITY
				sql.Clear();
				sql.Stmt = string.Format("INSERT INTO {0} (userId, reportId, readonly_flag) VALUES (@userId, {1}, 'false'); ", DbTables.RptAccessUser, id);
				sql.SqlParams.Add(new SqlParameter("@userId", sessionInfo.UserId));
				sql.ExecuteNQCmd();
			} catch (Exception e)
			{
				return e.Message;
			}
			return null;
		}

		/*
		 *  Generic method for updating a variety of report objects in the Reports table
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param options is a Dictionary<string, object> object containing parameters necessary for the record update
		 *  @return a string error message or, if successful, null
		 */
		private static string UpdateRptItem(SessionInfo sessionInfo, Dictionary<string, object> options)
		{
			try
			{ 
				SQLHandler sql = new SQLHandler();
				sql.Stmt = string.Format("UPDATE {0} SET {1}, " +
					"dateLastModified=getDate(), " +
					"lastModifiedByUserId=@userId " +
					"WHERE {2}", DbTables.Reports, options["change"], options["where"]);
				sql.SqlParams.Add(new SqlParameter("@userId", sessionInfo.UserId));
				sql.ExecuteNQCmd();
			}
			catch (Exception e)
			{
				return e.Message;
			}
			return null;
		}

		/*
		 *  Generic method for getting a variety of report objects from the Reports table
		 *  @param sessionInfo is a SessionInfo object as provided by Folder Management
		 *  @param options is a Dictionary<string, object> object containing parameters necessary for the record update
		 *  @return an object representation of what was requested
		 */
		private static object GetRptItem(SessionInfo sessionInfo, Dictionary<string, object> options)
		{
			bool isList = ((string)options["format"] == "list");	// querying list of items instead of a single item
			bool isRptOrFldr = options.ContainsKey("parent");       // querying a report or folder item or list
			bool isTheme = options.ContainsKey("themeType");		// querying a theme item or list

			StringBuilder sb = new StringBuilder();
			sb.Append(
				"SELECT {6} FROM {0} " +
				"JOIN {1} ON {0}.id={1}.reportId " +
				"WHERE {1}.userId=@userId " +
				"AND {0}.isDeleted='false' " +
				"AND {0}.type={5} ");

			if (!isList)
				sb.Append("AND {0}.name=@name ");
			if (isTheme)
				sb.Append("AND {0}.themeType=@themeType " );
			if (isRptOrFldr)
			{
				sb.Append("AND {0}.parentId " + options["parent"] + " ");
				sb.Append("AND {0}.leaf_flag='" + options["leaf"] + "' ");
			}

			sb.Append(
				"UNION " +
				"SELECT {6} FROM {0} " +
				"JOIN {2} ON {0}.id={2}.reportId " +
				"JOIN {4} ON {2}.userClassId={4}.classId " +
				"WHERE {4}.userId=@userId " +
				"AND {0}.isDeleted='false' " +
				"AND {0}.type={5} ");

			if (!isList)
				sb.Append("AND {0}.name=@name ");
			if (isTheme)
				sb.Append("AND {0}.themeType=@themeType ");
			if (isRptOrFldr)
			{
				sb.Append("AND {0}.parentId " + options["parent"] + " ");
				sb.Append("AND {0}.leaf_flag='" + options["leaf"] + "' ");
			}

			sb.Append(
				"UNION " +
				"SELECT {6} FROM {0} " +
				"JOIN {3} ON {0}.id={3}.reportId " +
				"WHERE {0}.isDeleted='false' " +
				"AND {0}.type={5} ");

			if (!isList)
				sb.Append("AND {0}.name=@name ");
			if (isTheme)
				sb.Append("AND {0}.themeType=@themeType ");
			if (isRptOrFldr)
			{
				sb.Append("AND {0}.parentId " + options["parent"] + " ");
				sb.Append("AND {0}.leaf_flag='" + options["leaf"] + "' ");
			}

			SQLHandler sql = new SQLHandler();
			sql.Stmt = string.Format(sb.ToString(),	DbTables.Reports, DbTables.RptAccessUser, DbTables.RptAccessUserClasses, 
				DbTables.RptPublished, DbTables.UserSettings, options["type"], options["whatToGet"]);

			if (!isList)
				sql.SqlParams.Add(new SqlParameter("@name", options["name"]));
			if (isTheme)
				sql.SqlParams.Add(new SqlParameter("@themeType", options["themeType"]));

			sql.SqlParams.Add(new SqlParameter("@userId", sessionInfo.UserId));
			DataSet ds = sql.ExecuteReadCmd();

			if (isList)
			{
				List<string> retVal = new List<string>();
				foreach (DataRow r in ds.Tables[0].Rows)
					retVal.Add(r[(string)options["whatToGet"]].ToString());

				return retVal;
			}
			else
				if (ds.Tables.Count == 0 || ds.Tables[0].Rows.Count == 0)
					return null;
				else
					return ds.Tables[0].Rows[0][(string)options["whatToGet"]];
		}
	}
}
