﻿using System;
using System.Data;
using System.Configuration;
using System.Data.SqlClient;
using System.Collections.Generic;

namespace DemoAppSQLUtils
{
	/// <summary>
	/// Wrapper class to facilitate the various types of database queries
	/// </summary>
	public class SQLHandler
	{
		private SqlConnection sqlConn = null;

		public string Stmt = "";											// SQL statement lives here
		public List<SqlParameter> SqlParams = new List<SqlParameter>();     // Non-string-format parameters live here
		public List<string> SqlTxnCmds = new List<string>();		// SQL commands to be executed as a transaction

		/**
		 * When a new SQLHandler is instantiated, initialize and open the connection.
		 */
		public SQLHandler()
		{
			InitConnection();
		}

		/**
		 * In an effort to reuse a SQLHandler object, this method allows for clearning out a previous SQL
		 * statement so that it can be used again.
		 */
		public void Clear()
		{
			Stmt = "";
			SqlParams.Clear();
		}

		/**
		 * Initialize the connection to the database.  Connection string lives in the main project's 
		 * web.config via standard configuration.  If SqlConnection is created successfully or pre-
		 * exists, open the connection.  If failure of any sort, clean up the SqlConnection and 
		 * re-throw the error to be managed at a higher level.
		 */ 
		private void InitConnection()
		{
			try
			{
				// if the SqlConnection object has not been disposed of, reuse it
				if (sqlConn == null)
					sqlConn = new SqlConnection(ConfigurationManager.ConnectionStrings["DefaultConnection"].ConnectionString);

				// once the SqlConnection has been established, if it's not open already, open it
				if (sqlConn != null && sqlConn.State == ConnectionState.Closed)
					sqlConn.Open();
			}
			catch
			{
				// if there was a problem, be sure to clean things up before we leave
				if (sqlConn != null)
					sqlConn.Dispose();

				throw;	// re-throw any exceptions to be managed at a higher level
			}
		}

		/**
		 * Executes a non-query SQL statement (meant for use with INSERT or UPDATE statements).  There
		 * is an assumption that a Stmt and, possibly, SqlParams are set prior to calling this method.  If not,
		 * an exception will be thrown.
		 */
		public void ExecuteNQCmd()
		{
			try
			{
				InitConnection();	// initialize database connection
				SqlCommand command = new SqlCommand(Stmt, sqlConn);

				// apply parameters, if any set
				foreach (SqlParameter p in SqlParams)
					command.Parameters.AddWithValue(p.ParameterName, p.Value);

				// execute sql statement
				command.ExecuteNonQuery();
			}
			catch
			{
				throw;	//re-throw any exceptions to be managed at a higher level
			}
			finally
			{
				// after all is said and done, close the database connection to keep things copacetic
				if (sqlConn != null)
					sqlConn.Close();
			}
		}

		/**
		 * Executes a SQL query (meant for use with SELECT statements returning a single value).  There
		 * is an assumption that a Stmt and, possibly, SqlParams are set prior to calling this method.  If not,
		 * an exception will be thrown.
		 * @return the queried object
		 */ 
		public Object ExecuteScalarCmd()
		{
			object retVal = null;

			try
			{
				InitConnection();	// initialize database connection
				SqlCommand command = new SqlCommand(Stmt, sqlConn);

				// apply parameters, if any set
				foreach (SqlParameter p in SqlParams)
					command.Parameters.AddWithValue(p.ParameterName, p.Value);

				// execute sql statement
				retVal = command.ExecuteScalar();
			}
			catch
			{
				throw;  //re-throw any exceptions to be managed at a higher level
			}
			finally
			{
				// after all is said and done, close the database connection to keep things copacetic
				if (sqlConn != null)
					sqlConn.Close();
			}
			return retVal;	// return queried object
		}

		/**
		 * Executes a SQL query (meant for use with SELECT statements returning a full DataSet object).  There
		 * is an assumption that a Stmt and, possibly, SqlParams are set prior to calling this method.  If not,
		 * an exception will be thrown.
		 * @return The resulting query's DataSet object
		 */ 
		public DataSet ExecuteReadCmd()
		{
			DataSet retVal = new DataSet();

			try
			{
				InitConnection();   // initialize database connection
				SqlDataAdapter adapter = new SqlDataAdapter(Stmt, sqlConn);

				// apply parameters, if any set
				foreach (SqlParameter p in SqlParams)
					adapter.SelectCommand.Parameters.AddWithValue(p.ParameterName, p.Value);

				// execute sql statement
				adapter.Fill(retVal);
			}
			catch
			{
				throw;  //re-throw any exceptions to be managed at a higher level
			}
			finally
			{
				// after all is said and done, close the database connection to keep things copacetic
				if (sqlConn != null)
					sqlConn.Close();
			}
			return retVal;  // return queried DataSet object
		}

		public void ExecTransaction()
		{
			SqlTransaction sqlTrans = null;
			try
			{
				InitConnection();   // initialize database connection

				sqlTrans = sqlConn.BeginTransaction();

				SqlCommand command = sqlConn.CreateCommand();
				command.Transaction = sqlTrans;

				// apply parameters, if any set; since we're using the same command, we only need to apply these once
				foreach (SqlParameter p in SqlParams)
					command.Parameters.AddWithValue(p.ParameterName, p.Value);

				foreach (string cmd in SqlTxnCmds)
				{
					command.CommandText = cmd;
					command.ExecuteNonQuery();
				}

				sqlTrans.Commit();					// commit the transaction
			}
			catch (Exception e)
			{
				try
				{
					sqlTrans.Rollback();
				}
				catch (Exception r)
				{
					throw;
				}

				throw;
			}
			finally
			{
				// after all is said and done, close the database connection to keep things copacetic
				if (sqlConn != null)
					sqlConn.Close();
			}
		}
	}

	/// <summary>
	/// Class that reads specific database table name appSettings configurations from the executing
	/// program's web.config and caches them in memory for fast reference.
	/// </summary>
	public static class DbTables
	{
		public static string Executions { get; set; }
		public static string Logins { get; set; }
		public static string DataObjPublished { get; set; }
		public static string DataObjUser { get; set; }
		public static string DataObjUserClasses { get; set; }
		public static string RptAccessUser { get; set; }
		public static string RptAccessUserClasses { get; set; }
		public static string RptPublished { get; set; }
		public static string Reports { get; set; }
		public static string UserPreferences { get; set; }
		public static string Companies { get; set; }
		public static string UserClasses { get; set; }
		public static string UserSettings { get; set; }
		public static string AspNetUsers { get; set; }

		static DbTables()
		{
			Executions = ConfigurationManager.AppSettings["tnExecutions"];
			Logins = ConfigurationManager.AppSettings["tnLogins"];
			DataObjPublished = ConfigurationManager.AppSettings["tnDataObjPublished"];
			DataObjUser = ConfigurationManager.AppSettings["tnDataObjUser"];
			DataObjUserClasses = ConfigurationManager.AppSettings["tnDataObjUserClasses"];
			RptAccessUser = ConfigurationManager.AppSettings["tnRptAccessUser"];
			RptAccessUserClasses = ConfigurationManager.AppSettings["tnRptAccessUserClasses"];
			RptPublished = ConfigurationManager.AppSettings["tnRptPublished"];
			Reports = ConfigurationManager.AppSettings["tnReports"];
			UserPreferences = ConfigurationManager.AppSettings["tnUserPreferences"];
			Companies = ConfigurationManager.AppSettings["tnCompanies"];
			UserClasses = ConfigurationManager.AppSettings["tnUserClasses"];
			UserSettings = ConfigurationManager.AppSettings["tnUserSettings"];
			AspNetUsers = ConfigurationManager.AppSettings["tnAspNetUsers"];
		}
	}
}